# Generated by Django 3.1.5 on 2021-03-09 11:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0089_auto_20210125_1649'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='timeentry',
            options={'ordering': ('-start_date_time', '-date_worked', '-id'), 'verbose_name_plural': 'Time entries'},
        ),
    ]
