# Generated by Django 2.1.14 on 2019-11-25 09:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0025_auto_20191125_0939'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='priority',
            options={'verbose_name_plural': 'Priorities'},
        ),
    ]
