# Generated by Django 2.1.14 on 2020-02-27 17:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0048_auto_20200211_1037'),
        ('djautotask', '0047_task_assigned_resource_role'),
    ]

    operations = [
    ]
