# Generated by Django 2.1.14 on 2019-11-26 16:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0028_phase'),
    ]

    operations = [
        migrations.AddField(
            model_name='phase',
            name='last_activity_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]
