# Generated by Django 2.1.14 on 2020-03-16 16:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0051_contract'),
    ]

    operations = [
        migrations.AddField(
            model_name='contract',
            name='status',
            field=models.CharField(blank=True, choices=[(0, 'Inactive'), (1, 'Active')], max_length=20, null=True),
        ),
    ]
